# Паралелен map с нишки

from threading import Thread, Lock
from queue import Queue


def fib(n):
    return fib(n - 2) + fib(n - 1) if n > 2 else n

# вариант 1
# - ненареден: редът на резултатите не е определен
# - по една нишка за всеки елемент в редицата (лошо...)
# - резултатите се връщат през синхронизирана опашкa
def pmap_unord(f, seq):
    a, q = [], Queue()
    def callf(x):
        q.put(f(x))
    n = 0
    for x in seq:
        th = Thread(target=callf, args=(x,))
        th.daemon = True
        th.start()
        n += 1
    for i in range(n):
        a.append(q.get())
    return a

# вариант 2
# - нареден: редът на резултатите съответства на елементите в редицата
def pmap_clumsy(f, seq):
    a = [None] * len(seq) # списък за всички резултати
    q = Queue()
    def callf(i, x):
        q.put((i, f(x)))
    for i, x in enumerate(seq):
        th = Thread(target=callf, args=(i, x))
        th.daemon = True
        th.start()
    for j in range(len(seq)):
        i, y = q.get()
        a[i] = y
    return a


# вариант 3:
# - има максимален брой на паралелните нишки
# - всяка нишка чака „задача“ от опашката и добавя резултат към спсъка с резутати
# - задачите са от вида (номер, аргумент)
# - пресметнатите стойности са от вида (номер, f(аргуемент))
# - накрая списъка със пресметнати стойности се подрежда по номер
def pmap(n, f, seq):
    a, alock, q = [], Lock(), Queue(n) # n - максимален брой елементи в опашката
    def worker():    # ако е достигнат .put() блокира докато не се извика .get()
        while True:
            t = q.get()
            if t is None: # с None сигнализираме за край
                return
            i, x = t
            with alock:
                a.append((i, f(x)))
            q.task_done()
    ws = [Thread(target=worker) for i in range(n)]
    for w in ws:
        w.start()
    for i, x in enumerate(seq):
        q.put((i, x))
    q.join() # чакаме да се пресметнат всички стойности
    for w in ws:
        q.put(None)
    for w in ws:
        w.join()
    a.sort(key=lambda t: t[0])
    for i in range(len(a)):
        a[i] = a[i][1]
    return a
